////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Plug-in defining a replicating resource. This resource makes sure that all of its data is replicated to all of its children
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// =-=-=-=-=-=-=-
// irods includes
#include "icatHighLevelRoutines.hpp"
#include "msParam.h"
#include "reGlobalsExtern.hpp"
#include "rodsLog.h"

// =-=-=-=-=-=-=-
#include "irods_resource_plugin.hpp"
#include "irods_file_object.hpp"
#include "irods_collection_object.hpp"
#include "irods_string_tokenize.hpp"
#include "irods_hierarchy_parser.hpp"
#include "irods_resource_backport.hpp"
#include "irods_plugin_base.hpp"
#include "irods_stacktrace.hpp"
#include "irods_repl_retry.hpp"
#include "irods_repl_types.hpp"
#include "irods_object_oper.hpp"
#include "irods_replicator.hpp"
#include "irods_create_write_replicator.hpp"
#include "irods_hierarchy_parser.hpp"
#include "irods_resource_redirect.hpp"
#include "irods_repl_rebalance.hpp"
#include "irods_kvp_string_parser.hpp"

// =-=-=-=-=-=-=-
// stl includes
#include <iostream>
#include <sstream>
#include <vector>
#include <string>
#include <map>
#include <list>
#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>

// =-=-=-=-=-=-=-
// system includes
#ifndef _WIN32
#include <sys/file.h>
#include <sys/param.h>
#endif
#include <errno.h>
#include <sys/stat.h>
#include <string.h>
#ifndef _WIN32
#include <unistd.h>
#endif
#include <sys/types.h>
#if defined(osx_platform)
#include <sys/malloc.h>
#else
#include <malloc.h>
#endif
#include <fcntl.h>
#ifndef _WIN32
#include <sys/file.h>
#include <unistd.h>
#endif
#include <dirent.h>

#if defined(solaris_platform)
#include <sys/statvfs.h>
#endif
#if defined(linux_platform)
#include <sys/vfs.h>
#endif
#include <sys/stat.h>

#include <string.h>

/// @brief Check the general parameters passed in to most plugin functions
template< typename DEST_TYPE >
irods::error replCheckParams(
    irods::resource_plugin_context& _ctx ) {
    irods::error result = SUCCESS();
    // =-=-=-=-=-=-=-
    // verify that the resc context is valid
    irods::error ret = _ctx.valid< DEST_TYPE >();
    if ( !ret.ok() ) {
        result = PASSMSG( "resource context is invalid", ret );
    }

    return result;
}

extern "C" {
    /// =-=-=-=-=-=-=-
    /// @brief limit of the number of repls to operate upon during rebalance
    const int DEFAULT_LIMIT = 500;

    const std::string RECURSIVE_REBALANCE_FLAG("recursive_rebalance");

    // =-=-=-=-=-=-=-
    // 2. Define operations which will be called by the file*
    //    calls declared in server/driver/include/fileDriver.h
    // =-=-=-=-=-=-=-

    // =-=-=-=-=-=-=-
    // NOTE :: to access properties in the _prop_map do the
    //      :: following :
    //      :: double my_var = 0.0;
    //      :: irods::error ret = _prop_map.get< double >( "my_key", my_var );
    // =-=-=-=-=-=-=-

    //////////////////////////////////////////////////////////////////////
    // Utility functions

    /**
     * @brief Gets the name of the child of this resource from the hierarchy
     */
    irods::error replGetNextRescInHier(
        const irods::hierarchy_parser& _parser,
        irods::resource_plugin_context& _ctx,
        irods::resource_ptr& _ret_resc ) {
        irods::error result = SUCCESS();
        irods::error ret;
        std::string this_name;
        ret = _ctx.prop_map().get<std::string>( irods::RESOURCE_NAME, this_name );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - Failed to get resource name from property map.";
            result = ERROR( -1, msg.str() );
        }
        else {
            std::string child;
            ret = _parser.next( this_name, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in the hierarchy.";
                result = ERROR( -1, msg.str() );
            }
            else {
                _ret_resc = ( _ctx.child_map() )[child].second;
            }
        }
        return result;
    }

    /// @brief Returns true if the specified object is in the specified object list
    bool replObjectInList(
        const object_list_t& _object_list,
        const irods::file_object_ptr _object,
        irods::object_oper& _rtn_oper ) {
        bool result = false;
        object_list_t::const_iterator it;
        for ( it = _object_list.begin(); !result && it != _object_list.end(); ++it ) {
            irods::object_oper oper = *it;
            if ( oper.object() == ( *_object.get() ) ) {
                _rtn_oper = oper;
                result = true;
            }
        }
        return result;
    }

    /// @brief Updates the fields in the resources properties for the object
    irods::error replUpdateObjectAndOperProperties(
        irods::resource_plugin_context& _ctx,
        const std::string& _oper ) {
        irods::error result = SUCCESS();
        irods::error ret;
        object_list_t object_list;
        // The object list is now a queue of operations and their associated objects. Their corresponding replicating operations
        // will be performed one at a time in the order in which they were put into the queue.
        irods::file_object_ptr file_obj = boost::dynamic_pointer_cast<irods::file_object >( ( _ctx.fco() ) );
        ret = _ctx.prop_map().get<object_list_t>( OBJECT_LIST_PROP, object_list );
        irods::object_oper oper;
        if ( !ret.ok() && ret.code() != KEY_NOT_FOUND ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - Failed to get the object list from the resource.";
            result = PASSMSG( msg.str(), ret );
        }
        else if ( replObjectInList( object_list, file_obj, oper ) ) {
            // confirm the operations are compatible
            bool mismatched = false;
            if ( _oper == irods::CREATE_OPERATION ) {
                if ( oper.operation() != irods::CREATE_OPERATION ) {
                    mismatched = true;
                }
            }
            else if ( _oper == irods::WRITE_OPERATION ) {
                // write is allowed after create
                if ( oper.operation() != irods::CREATE_OPERATION && oper.operation() != irods::WRITE_OPERATION ) {
                    mismatched = true;
                }
            }

            result = ASSERT_ERROR( !mismatched, INVALID_OPERATION,
                                   "Existing object operation: \"%s\" does not match current operation: \"%s\".",
                                   oper.operation().c_str(), _oper.c_str() );
        }
        else {
            oper.object() = *( file_obj.get() );
            oper.operation() = _oper;
            object_list.push_back( oper );
            ret = _ctx.prop_map().set<object_list_t>( OBJECT_LIST_PROP, object_list );
            result = ASSERT_PASS( ret, "Failed to set the object list property on the resource." );
        }

        if ( !result.ok() ) {
            irods::log( result );
        }

        return result;
    }

    bool is_self_in_hier(
        irods::resource_plugin_context& _ctx,
        const irods::hierarchy_parser &_parser) {
        std::string name;
        irods::error ret = _ctx.prop_map().get<std::string>(irods::RESOURCE_NAME, name);
        if (!ret.ok()) {
            THROW(ret.code(), ret.result());
        }
        return _parser.resc_in_hier(name);
    }

    irods::error get_selected_hierarchy(
        irods::resource_plugin_context& _ctx,
        std::string& _hier_string,
        std::string& _root_resc ) {
        irods::hierarchy_parser selected_parser;
        std::string operation;

        try {
            bool resc_hier_in_kw = false;
            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast<irods::file_object >( _ctx.fco() );
            if (file_obj->l1_desc_idx() > 0) {
                char* hier_str = getValByKey(&L1desc[file_obj->l1_desc_idx()].dataObjInp->condInput, RESC_HIER_STR_KW);
                if (!hier_str) {
                    return ERROR(SYS_INTERNAL_NULL_INPUT_ERR,
                                 (boost::format(
                                  "[%s] - No hierarchy string found in keywords for file object.") %
                                  __FUNCTION__).str().c_str());
                }
                selected_parser.set_string(hier_str);
                resc_hier_in_kw = is_self_in_hier(_ctx, selected_parser);
            }
            if (!resc_hier_in_kw) {
                std::string selected_hier = file_obj->resc_hier();
                if (selected_hier.empty()) {
                    return ERROR(SYS_INTERNAL_NULL_INPUT_ERR,
                                 (boost::format(
                                  "[%s] - file object does not have a resource hierarchy.") %
                                  __FUNCTION__).str().c_str());
                }
                selected_parser.set_string(selected_hier);
                if (!is_self_in_hier(_ctx, selected_parser)) {
                    return ERROR(HIERARCHY_ERROR,
                                 (boost::format(
                                  "[%s] - Replicating a file object which does not exist in this hierarchy.") %
                                  __FUNCTION__).str().c_str());
                }
            }
        }
        catch (const irods::exception &e) {
            return irods::error(e);
        }

        irods::error ret = selected_parser.str( _hier_string );
        if ( !ret.ok() ) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to get the hierarchy string from the parser.") %
                        __FUNCTION__).str(), ret );
        }

        ret = selected_parser.first_resc( _root_resc );
        if ( !ret.ok() ) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to get the root resource from the parser.") %
                        __FUNCTION__).str(), ret );
        }

        return SUCCESS();
    }

    irods::error replReplicateCreateWrite(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();
        // get the list of objects that need to be replicated
        object_list_t object_list;
        irods::error ret = _ctx.prop_map().get<object_list_t>( OBJECT_LIST_PROP, object_list );
        if (!ret.ok() && KEY_NOT_FOUND != ret.code()) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to get object list for replication.") %
                        __FUNCTION__).str(), ret );
        }

        if ( object_list.empty() ) {
            return SUCCESS();
        }

        // get the child list
        child_list_t child_list;
        ret = _ctx.prop_map().get<child_list_t>( CHILD_LIST_PROP, child_list );
        if ( !ret.ok() ) {
            irods::log(PASSMSG( (boost::format(
                       "[%s] - Failed to get child list for replication.") %
                       __FUNCTION__).str(), ret ));
            return SUCCESS();
        }

        // get the root resource name as well as the child hierarchy string
        std::string root_resc;
        std::string child;
        ret = get_selected_hierarchy( _ctx, child, root_resc );
        if ( !ret.ok() ) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to determine the root resource and selected hierarchy.") %
                        __FUNCTION__).str(), ret );
        }

        std::string name;
        ret = _ctx.prop_map().get<std::string>( irods::RESOURCE_NAME, name );
        if ( !ret.ok() ) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Could not determine resource name.") %
                        __FUNCTION__).str(), ret );
        }

        irods::create_write_replicator oper_repl( root_resc, name, child );
        irods::replicator replicator( &oper_repl );
        ret = replicator.replicate( _ctx, child_list, object_list );
        if ( !ret.ok() ) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to replicate the create/write operation to the siblings.") %
                        __FUNCTION__).str(), ret );
        }

        // update the object list in the properties
        ret = _ctx.prop_map().set<object_list_t>( OBJECT_LIST_PROP, object_list );
        if ( !ret.ok() ) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to update the object list in the properties.") %
                        __FUNCTION__).str(), ret );
        }
        return SUCCESS();
    }

    //////////////////////////////////////////////////////////////////////
    // Actual operations

    // Called after a new file is registered with the ICAT
    irods::error replFileRegistered(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();
        irods::error ret;
        ret = replCheckParams< irods::file_object >( _ctx );
        if ( ( result = ASSERT_PASS( ret, "Error checking passed paramters." ) ).ok() ) {

            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast<irods::file_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( ( result = ASSERT_PASS( ret, "Failed to get the next resource in hierarchy." ) ).ok() ) {
                ret = child->call( _ctx.comm(), irods::RESOURCE_OP_REGISTERED, _ctx.fco() );
                result = ASSERT_PASS( ret, "Failed while calling child operation." );
            }
        }
        return result;
    }

    // Called when a file is unregistered from the ICAT
    irods::error replFileUnregistered(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();
        irods::error ret;
        ret = replCheckParams< irods::file_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - Error found checking passed parameters.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast< irods::file_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call( file_obj->comm(), irods::RESOURCE_OP_UNREGISTERED, file_obj ) ;
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
            }
        }
        return result;
    }

    irods::error create_replication_list(
        irods::resource_plugin_context& _ctx,
        const std::string& operation) {

        // Get selected hierarchy string
        std::string selected_hier;
        std::string root_resc;
        irods::error ret = get_selected_hierarchy(_ctx, selected_hier, root_resc);
        if (!ret.ok()) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to get selected hierarchy.") %
                        __FUNCTION__).str(), ret);
        }

        // Get resource name and chop down parser until hierarchy represent this resource
        irods::hierarchy_parser selected_hier_parser;
        selected_hier_parser.set_string(selected_hier);
        std::string name;
        ret = _ctx.prop_map().get<std::string>(irods::RESOURCE_NAME, name);
        if (!ret.ok()) {
            return PASS(ret);
        }
        std::string current_resc_hier;
        ret = selected_hier_parser.str(current_resc_hier, name);
        if (!ret.ok()) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to get hierarchy string for resc name [%s].") %
                        __FUNCTION__ % name).str(), ret);
        }
        irods::hierarchy_parser current_hier_parser;
        current_hier_parser.set_string(current_resc_hier);

        // Get current hostname for vote
        char host_name_str[MAX_NAME_LEN];
        if (gethostname( host_name_str, MAX_NAME_LEN ) < 0) {
            return ERROR(SYS_GET_HOSTNAME_ERR, "failed in gethostname");
        }
        const std::string host_name(host_name_str);

        // Loop over children and create a list of hierarchies to which objects will be replicated
        float out_vote;
        child_list_t repl_vector;
        irods::resource_child_map::iterator it;
        for ( it = _ctx.child_map().begin(); it != _ctx.child_map().end(); ++it ) {
            irods::resource_ptr child = it->second.second;

            // Get child resource name and skip resolution if selected hierarchy descends from this child
            std::string child_name;
            ret = child->get_property<std::string>(irods::RESOURCE_NAME, child_name);
            if (!ret.ok()) {
                return PASSMSG(
                           (boost::format(
                            "[%s] - Failed to get resource name for child.") %
                            __FUNCTION__).str(), ret );
            }
            if (selected_hier_parser.resc_in_hier(child_name)) {
                continue;
            }

            irods::hierarchy_parser parser( current_hier_parser );
            irods::error ret = child->call<const std::string*, const std::string*, irods::hierarchy_parser*, float*>(
                      _ctx.comm(), irods::RESOURCE_OP_RESOLVE_RESC_HIER, _ctx.fco(), &operation, &host_name, &parser, &out_vote );
            if ( !ret.ok() && ret.code() != CHILD_NOT_FOUND ) {
                irods::log(PASSMSG((boost::format(
                                    "[%s] - Failed calling redirect on the child \"%s\".") %
                                    __FUNCTION__ % it->first).str(), ret ));
            }
            else if ( out_vote > 0 ) {
                repl_vector.push_back(parser);
            }
        }

        // add the resulting vector as a property of the resource
        ret = _ctx.prop_map().set<child_list_t>(CHILD_LIST_PROP, repl_vector);
        if ( !ret.ok() ) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to store the repl child list as a property.") %
                        __FUNCTION__).str(), ret );
        }

        return SUCCESS();

    } // create_replication_list

    // Called when a files entry is modified in the ICAT
    irods::error replFileModified(
        irods::resource_plugin_context& _ctx ) {

        irods::error ret = replCheckParams<irods::file_object>(_ctx);
        if (!ret.ok()) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Error checking passed parameters.") %
                        __FUNCTION__).str(), ret);
        }

        // Get next resource on which to call file_modified
        irods::file_object_ptr file_obj = boost::dynamic_pointer_cast<irods::file_object>(_ctx.fco());
        irods::hierarchy_parser parser;
        parser.set_string(file_obj->resc_hier());
        irods::resource_ptr child;
        ret = replGetNextRescInHier(parser, _ctx, child);
        if (!ret.ok()) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to get the next resource in hierarchy.") %
                        __FUNCTION__).str(), ret);
        }

        // Call file_modified on next resource in hierarchy
        ret = child->call( _ctx.comm(), irods::RESOURCE_OP_MODIFIED, _ctx.fco() );
        if (!ret.ok()) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed while calling child operation.") %
                        __FUNCTION__).str(), ret);
        }

        std::string name;
        ret = _ctx.prop_map().get<std::string>(irods::RESOURCE_NAME, name);
        if (!ret.ok()) {
            return PASS(ret);
        }

        irods::hierarchy_parser sub_parser;
        sub_parser.set_string(file_obj->in_pdmo());
        if (!sub_parser.resc_in_hier(name)) {
            std::string operation;
            const char* open_type_str = getValByKey( &file_obj->cond_input(), OPEN_TYPE_KW );
            if (!open_type_str) {
                // If open type is not set, return early to avoid replication
                return SUCCESS();
            }

            const int open_type = atoi(open_type_str);
            if (CREATE_TYPE == open_type) {
                operation = irods::CREATE_OPERATION;
            }
            else if (OPEN_FOR_WRITE_TYPE == open_type) {
                operation = irods::WRITE_OPERATION;
            }
            else {
                return ERROR(SYS_INVALID_INPUT_PARAM, "openType not valid for replication.");
            }

            ret = create_replication_list(_ctx, operation);
            if (!ret.ok()) {
                return PASS(ret);
            }

            ret = replUpdateObjectAndOperProperties(_ctx, operation);
            if (!ret.ok()) {
                return PASS(ret);
            }

            ret = replReplicateCreateWrite( _ctx );
            if (!ret.ok()) {
                return PASSMSG((boost::format(
                            "[%s] - Failed to replicate create/write operation for object: \"%s\".") %
                            __FUNCTION__ % file_obj->logical_path()).str(), ret);
            }
        }

        return SUCCESS();
    } // replFileModified

    // =-=-=-=-=-=-=-
    // interface for POSIX create
    irods::error replFileCreate(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();
        irods::error ret;

        // get the list of objects that need to be replicated
        object_list_t object_list;
        ret = _ctx.prop_map().get<object_list_t>( OBJECT_LIST_PROP, object_list );

        ret = replCheckParams< irods::file_object >( _ctx );
        if ( !ret.ok() ) {
            result = PASSMSG( "replFileCreatePlugin - bad params.", ret );
        }
        else {
            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast< irods::file_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call( _ctx.comm(), irods::RESOURCE_OP_CREATE, _ctx.fco() );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
            }
        }
        return result;
    } // replFileCreate

    // =-=-=-=-=-=-=-
    // interface for POSIX Open
    irods::error replFileOpen(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::file_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast<irods::file_object >( ( _ctx.fco() ) );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call( _ctx.comm(), irods::RESOURCE_OP_OPEN, _ctx.fco() );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
            }
        }
        return result;
    } // replFileOpen

    // =-=-=-=-=-=-=-
    // interface for POSIX Read
    irods::error replFileRead(
        irods::resource_plugin_context& _ctx,
        void*                          _buf,
        int                            _len ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::file_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast<irods::file_object >( ( _ctx.fco() ) );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call<void*, int>( _ctx.comm(), irods::RESOURCE_OP_READ, _ctx.fco(), _buf, _len );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    // have to return the actual code because it contains the number of bytes read
                    result = CODE( ret.code() );
                }
            }
        }
        return result;
    } // replFileRead

    // =-=-=-=-=-=-=-
    // interface for POSIX Write
    irods::error replFileWrite(
        irods::resource_plugin_context& _ctx,
        void*                          _buf,
        int                            _len ) {
        irods::error result = SUCCESS();
        irods::error ret;
        // get the list of objects that need to be replicated
        object_list_t object_list;
        ret = _ctx.prop_map().get<object_list_t>( OBJECT_LIST_PROP, object_list );

        ret = replCheckParams< irods::file_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast<irods::file_object >( ( _ctx.fco() ) );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call<void*, int>( _ctx.comm(), irods::RESOURCE_OP_WRITE, _ctx.fco(), _buf, _len );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );

                }
            }
        }
        return result;
    } // replFileWrite

    // =-=-=-=-=-=-=-
    // interface for POSIX Close
    irods::error replFileClose(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();
        irods::error ret;
        // get the list of objects that need to be replicated
        object_list_t object_list;
        ret = _ctx.prop_map().get<object_list_t>( OBJECT_LIST_PROP, object_list );

        ret = replCheckParams< irods::file_object >( _ctx );
        if ( ( result = ASSERT_PASS( ret, "Bad params." ) ).ok() ) {

            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast<irods::file_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( ( result = ASSERT_PASS( ret, "Failed to get the next resource in hierarchy." ) ).ok() ) {

                ret = child->call( _ctx.comm(), irods::RESOURCE_OP_CLOSE, _ctx.fco() );
                result = ASSERT_PASS( ret, "Failed while calling child operation." );
            }
        }
        return result;

    } // replFileClose

    // =-=-=-=-=-=-=-
    // interface for POSIX Unlink
    irods::error replFileUnlink(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::data_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::data_object_ptr data_obj = boost::dynamic_pointer_cast<irods::data_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( data_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call( _ctx.comm(), irods::RESOURCE_OP_UNLINK, _ctx.fco() );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );

                }
            }
        }
        return result;
    } // replFileUnlink

    // =-=-=-=-=-=-=-
    // interface for POSIX Stat
    irods::error replFileStat(
        irods::resource_plugin_context& _ctx,
        struct stat*                   _statbuf ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::data_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::data_object_ptr data_obj = boost::dynamic_pointer_cast< irods::data_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( data_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call<struct stat*>( _ctx.comm(), irods::RESOURCE_OP_STAT, _ctx.fco(), _statbuf );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );
                }
            }
        }
        return result;
    } // replFileStat

    // =-=-=-=-=-=-=-
    // interface for POSIX lseek
    irods::error replFileLseek(
        irods::resource_plugin_context& _ctx,
        size_t                         _offset,
        int                            _whence ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::file_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast< irods::file_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call<size_t, int>( _ctx.comm(), irods::RESOURCE_OP_LSEEK, _ctx.fco(), _offset, _whence );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );
                }
            }
        }
        return result;
    } // replFileLseek

    // =-=-=-=-=-=-=-
    // interface for POSIX mkdir
    irods::error replFileMkdir(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::collection_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::collection_object_ptr collection_obj = boost::dynamic_pointer_cast< irods::collection_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( collection_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call( _ctx.comm(), irods::RESOURCE_OP_MKDIR, _ctx.fco() );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );
                }
            }
        }
        return result;
    } // replFileMkdir

    // =-=-=-=-=-=-=-
    // interface for POSIX mkdir
    irods::error replFileRmdir(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::collection_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::collection_object_ptr file_obj = boost::dynamic_pointer_cast< irods::collection_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call( _ctx.comm(), irods::RESOURCE_OP_RMDIR, _ctx.fco() );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );
                }
            }
        }
        return result;
    } // replFileRmdir

    // =-=-=-=-=-=-=-
    // interface for POSIX opendir
    irods::error replFileOpendir(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::collection_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::collection_object_ptr collection_obj = boost::dynamic_pointer_cast< irods::collection_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( collection_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call( _ctx.comm(), irods::RESOURCE_OP_OPENDIR, _ctx.fco() );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );
                }
            }
        }
        return result;
    } // replFileOpendir

    // =-=-=-=-=-=-=-
    // interface for POSIX closedir
    irods::error replFileClosedir(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::collection_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::collection_object_ptr collection_obj = boost::dynamic_pointer_cast< irods::collection_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( collection_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call( _ctx.comm(), irods::RESOURCE_OP_CLOSEDIR, _ctx.fco() );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );
                }
            }
        }
        return result;
    } // replFileClosedir

    // =-=-=-=-=-=-=-
    // interface for POSIX readdir
    irods::error replFileReaddir(
        irods::resource_plugin_context& _ctx,
        struct rodsDirent**            _dirent_ptr ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::collection_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::collection_object_ptr collection_obj = boost::dynamic_pointer_cast< irods::collection_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( collection_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call<rodsDirent**>( _ctx.comm(), irods::RESOURCE_OP_READDIR, _ctx.fco(), _dirent_ptr );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );
                }
            }
        }
        return result;
    } // replFileReaddir

    // =-=-=-=-=-=-=-
    // interface for POSIX readdir
    irods::error replFileRename(
        irods::resource_plugin_context& _ctx,
        const char*                    _new_file_name ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::file_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast< irods::file_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call<const char*>( _ctx.comm(), irods::RESOURCE_OP_RENAME, _ctx.fco(), _new_file_name );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );
                }
            }
        }
        return result;
    } // replFileRename

    // =-=-=-=-=-=-=-
    // interface for POSIX truncate
    irods::error replFileTruncate(
        irods::resource_plugin_context& _ctx ) {
        // =-=-=-=-=-=-=-
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams<irods::file_object>( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::file_object_ptr ptr = boost::dynamic_pointer_cast< irods::file_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( ptr->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call( _ctx.comm(), irods::RESOURCE_OP_TRUNCATE, _ctx.fco() );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );
                }
            }
        }
        return result;
    } // replFileTruncate

    // =-=-=-=-=-=-=-
    // interface to determine free space on a device given a path
    irods::error replFileGetFsFreeSpace(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::file_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast< irods::file_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call( _ctx.comm(), irods::RESOURCE_OP_FREESPACE, _ctx.fco() );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );
                }
            }
        }
        return result;
    } // replFileGetFsFreeSpace

    // =-=-=-=-=-=-=-
    // replStageToCache - This routine is for testing the TEST_STAGE_FILE_TYPE.
    // Just copy the file from filename to cacheFilename. optionalInfo info
    // is not used.
    irods::error replStageToCache(
        irods::resource_plugin_context& _ctx,
        const char*                      _cache_file_name ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::file_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast< irods::file_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call< const char* >(
                          _ctx.comm(),
                          irods::RESOURCE_OP_STAGETOCACHE,
                          _ctx.fco(),
                          _cache_file_name );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
            }
        }
        return result;
    } // replStageToCache

    // =-=-=-=-=-=-=-
    // passthruSyncToArch - This routine is for testing the TEST_STAGE_FILE_TYPE.
    // Just copy the file from cacheFilename to filename. optionalInfo info
    // is not used.
    irods::error replSyncToArch(
        irods::resource_plugin_context& _ctx,
        const char*                    _cache_file_name ) {
        irods::error result = SUCCESS();
        irods::error ret;

        ret = replCheckParams< irods::file_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast< irods::file_object >( _ctx.fco() );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call<const char*>( _ctx.comm(), irods::RESOURCE_OP_SYNCTOARCH, _ctx.fco(), _cache_file_name );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
                else {
                    result = CODE( ret.code() );
                }
            }
        }
        return result;
    } // replSyncToArch

    /// @brief Adds the current resource to the specified resource hierarchy
    irods::error replAddSelfToHierarchy(
        irods::resource_plugin_context& _ctx,
        irods::hierarchy_parser& _parser ) {
        irods::error result = SUCCESS();
        irods::error ret;
        std::string name;
        ret = _ctx.prop_map().get<std::string>( irods::RESOURCE_NAME, name );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - Failed to get the resource name.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            ret = _parser.add_child( name );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to add resource to hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
        }
        return result;
    }

    /// @brief Loop through the children and call redirect on each one to populate the hierarchy vector
    irods::error replRedirectToChildren(
        irods::resource_plugin_context& _ctx,
        const std::string*             _operation,
        const std::string*             _curr_host,
        irods::hierarchy_parser&       _parser,
        redirect_map_t&                _redirect_map ) {
        irods::error result = SUCCESS();
        irods::error ret;
        irods::resource_child_map::iterator it;
        float out_vote;
        for ( it = _ctx.child_map().begin(); result.ok() && it != _ctx.child_map().end(); ++it ) {
            irods::hierarchy_parser parser( _parser );
            irods::resource_ptr child = it->second.second;
            ret = child->call<const std::string*, const std::string*, irods::hierarchy_parser*, float*>(
                      _ctx.comm(), irods::RESOURCE_OP_RESOLVE_RESC_HIER, _ctx.fco(), _operation, _curr_host, &parser, &out_vote );
            if ( !ret.ok() && ret.code() != CHILD_NOT_FOUND ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed calling redirect on the child \"" << it->first << "\"";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                _redirect_map.insert( std::pair<float, irods::hierarchy_parser>( out_vote, parser ) );
            }
        }
        return result;
    }

    /// @brief Selects a child from the vector of parsers based on host access
    irods::error replSelectChild(
        const redirect_map_t&           _redirect_map,
        irods::hierarchy_parser*        _out_parser,
        float*                          _out_vote ) {

        (*_out_vote) = 0.0;
        if( _redirect_map.empty() ) {
            // there are no votes to consider
            return SUCCESS();
        }

        redirect_map_t::const_iterator it;
        it = _redirect_map.begin();
        float vote = it->first;
        irods::hierarchy_parser parser = it->second;
        *_out_parser = parser;
        *_out_vote = vote;

        std::string hier;
        parser.str(hier);
        rodsLog(LOG_DEBUG, "[%s] - repl selected hier:[%s],vote:[%f]", __FUNCTION__, hier.c_str(), vote);
        if ( 0.0 == vote ) {
            return SUCCESS();
        }

        return SUCCESS();
    } // replSelectChild

    /// @brief Determines which child should be used for the specified operation
    irods::error replRedirect(
        irods::resource_plugin_context& _ctx,
        const std::string*              _operation,
        const std::string*              _curr_host,
        irods::hierarchy_parser*        _inout_parser,
        float*                          _out_vote ) {
        if (NULL == _operation || NULL == _curr_host || NULL == _inout_parser || NULL == _out_vote) {
            return ERROR( SYS_INVALID_INPUT_PARAM,
                          (boost::format(
                           "[%s]: null parameters passed to redirect") %
                           __FUNCTION__).str().c_str() ); 
        }

        // add ourselves to the hierarchy parser
        irods::hierarchy_parser parser = *_inout_parser;
        irods::error ret = replAddSelfToHierarchy(_ctx, parser);
        if (!ret.ok()) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to add ourselves to the resource hierarchy.") %
                        __FUNCTION__).str(), ret);
        }

        // Resolve each one of our children and put into redirect_map
        redirect_map_t redirect_map;
        ret = replRedirectToChildren(_ctx, _operation, _curr_host, parser, redirect_map);
        if (!ret.ok()) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to redirect to all children.") %
                        __FUNCTION__).str(), ret);
        }

        // Select a resolved hierarchy from redirect_map for the operation
        ret = replSelectChild(redirect_map, _inout_parser, _out_vote);
        if (!ret.ok()) {
            return PASSMSG(
                       (boost::format(
                        "[%s] - Failed to select an appropriate child.") %
                        __FUNCTION__).str(), ret);
        }

        return SUCCESS();
    } // replRedirect

    // =-=-=-=-=-=-=-
    // replRebalance - code which would rebalance the subtree
    irods::error replRebalance(
        irods::resource_plugin_context& _ctx ) {
        irods::error result = SUCCESS();

        std::string recurse_flag;
        irods::error ret = _ctx.prop_map().get<std::string>(
                               RECURSIVE_REBALANCE_FLAG,
                               recurse_flag );
        std::string lower_recurse_flag = boost::algorithm::to_lower_copy( recurse_flag );
        if(!ret.ok() || "false" != lower_recurse_flag) {
            // =-=-=-=-=-=-=-
            // forward request for rebalance to children first, then
            // rebalance ourselves.
            irods::resource_child_map::iterator itr = _ctx.child_map().begin();
            for ( ; itr != _ctx.child_map().end(); ++itr ) {
                irods::error ret = itr->second.second->call(
                                       _ctx.comm(),
                                       irods::RESOURCE_OP_REBALANCE,
                                       _ctx.fco() );
                if ( !ret.ok() ) {
                    irods::log( PASS( ret ) );
                    result = ret;
                }
            }
        }

        // =-=-=-=-=-=-=-
        // if our children had errors, then we bail
        if ( !result.ok() ) {
            return PASS( result );
        }

        // =-=-=-=-=-=-=-
        // get the property 'name' of this resource
        std::string resc_name;
        ret = _ctx.prop_map().get< std::string >(
                  irods::RESOURCE_NAME,
                  resc_name );
        if ( !ret.ok() ) {
            return PASS( ret );
        }

        // =-=-=-=-=-=-=-
        // determine limit size
        int limit = DEFAULT_LIMIT;
        if ( !_ctx.rule_results().empty() ) {
            irods::kvp_map_t kvp;
            irods::error kvp_err = irods::parse_kvp_string(
                                       _ctx.rule_results(),
                                       kvp );
            if ( !kvp_err.ok() ) {
                return PASS( kvp_err );
            }

            std::string limit_str = kvp[ irods::REPL_LIMIT_KEY ];
            if ( !limit_str.empty() ) {
                try {
                    limit = boost::lexical_cast<int>( limit_str );

                }
                catch ( const boost::bad_lexical_cast& ) {
                    std::stringstream msg;
                    msg << "failed to cast value ["
                        << limit_str
                        << "] to an integer";
                    return ERROR(
                               SYS_INVALID_INPUT_PARAM,
                               msg.str() );
                }
            }

        } // if rule results not empty

        // =-=-=-=-=-=-=-
        // determine distinct root count
        int status = 0;
        long long root_count = 0;
        status = chlGetDistinctDataObjCountOnResource(
                     resc_name.c_str(),
                     root_count );
        if ( status < 0 ) {
            std::stringstream msg;
            msg << "failed to get distinct cound for ["
                << resc_name
                << "]";
            return ERROR( status, msg.str().c_str() );
        }

        // =-=-=-=-=-=-=-
        // get invocation timestamp
        // only replicas at rest prior to this invocation timestamp will be rebalanced
        // prevents race condition of new/inflight data being 'over'-rebalanced
        char invocation_timestamp[50];
        getNowStr(invocation_timestamp);

        // =-=-=-=-=-=-=-
        // iterate over the children, if one does not have a matching
        // distinct count then we need to rebalance
        irods::resource_child_map::iterator c_itr;
        for ( c_itr  = _ctx.child_map().begin();
                c_itr != _ctx.child_map().end();
                ++c_itr ) {
            // =-=-=-=-=-=-=-
            // get list of distinct missing data ids or dirty repls, iterate
            // until query fails - no more repls necessary for child
            dist_child_result_t results( 1 );
            irods::ReasonForReplication reason_for_replication;
            while ( !results.empty() ) {
                irods::error ga_ret = irods::gather_data_objects_for_rebalance(
                                          _ctx.comm(),
                                          resc_name,
                                          c_itr->first,
                                          limit,
                                          invocation_timestamp,
                                          results,
                                          reason_for_replication);
                if ( ga_ret.ok() ) {
                    if ( !results.empty() ) {
                        // =-=-=-=-=-=-=-
                        // if the results are not empty call our processing function
                        irods::error proc_ret = irods::proc_results_for_rebalance(
                                                    _ctx,         // context for repl node
                                                    resc_name,    // parent resc name
                                                    c_itr->first, // child resc name
                                                    results,      // result set
                                                    reason_for_replication);
                        if ( !proc_ret.ok() ) {
                            return PASS( proc_ret );
                        }

                    } // if results

                }
                else {
                    return PASS( ga_ret );

                }

            } // while

        } // for c_itr

        return SUCCESS();

    } // replRebalance

    // Called when a files entry is modified in the ICAT
    irods::error replFileNotify(
        irods::resource_plugin_context& _ctx,
        const std::string*               _opr ) {
        irods::error result = SUCCESS();
        irods::error ret = replCheckParams< irods::file_object >( _ctx );
        if ( !ret.ok() ) {
            std::stringstream msg;
            msg << __FUNCTION__;
            msg << " - bad params.";
            result = PASSMSG( msg.str(), ret );
        }
        else {
            irods::file_object_ptr file_obj = boost::dynamic_pointer_cast<irods::file_object >( ( _ctx.fco() ) );
            irods::hierarchy_parser parser;
            parser.set_string( file_obj->resc_hier() );
            irods::resource_ptr child;
            ret = replGetNextRescInHier( parser, _ctx, child );
            if ( !ret.ok() ) {
                std::stringstream msg;
                msg << __FUNCTION__;
                msg << " - Failed to get the next resource in hierarchy.";
                result = PASSMSG( msg.str(), ret );
            }
            else {
                ret = child->call( _ctx.comm(), irods::RESOURCE_OP_NOTIFY, _ctx.fco(), _opr );
                if ( !ret.ok() ) {
                    std::stringstream msg;
                    msg << __FUNCTION__;
                    msg << " - Failed while calling child operation.";
                    result = PASSMSG( msg.str(), ret );
                }
            }
        }
        return result;
    }

    // =-=-=-=-=-=-=-
    // 3. create derived class to handle unix file system resources
    //    necessary to do custom parsing of the context string to place
    //    any useful values into the property map for reference in later
    //    operations.  semicolon is the preferred delimiter
    class repl_resource : public irods::resource {

        public:
            repl_resource(
                const std::string& _inst_name,
                const std::string& _context ) :
                irods::resource( _inst_name, _context ) {

                if ( _context.empty() ) {
                    // Retry capability requires default values to be used if not set
                    properties_.set< uint32_t >( irods::RETRY_ATTEMPTS_KW, irods::DEFAULT_RETRY_ATTEMPTS );
                    properties_.set< uint32_t >( irods::RETRY_FIRST_DELAY_IN_SECONDS_KW, irods::DEFAULT_RETRY_FIRST_DELAY_IN_SECONDS );
                    properties_.set< double >( irods::RETRY_BACKOFF_MULTIPLIER_KW, irods::DEFAULT_RETRY_BACKOFF_MULTIPLIER );
                    return;
                }

                // =-=-=-=-=-=-=-
                // parse context string into property pairs assuming a ; as a separator
                std::vector< std::string > props;
                irods::kvp_map_t kvp_map;
                irods::parse_kvp_string( _context, kvp_map );

                // =-=-=-=-=-=-=-
                // copy the properties from the context to the prop map
                irods::kvp_map_t::iterator itr = kvp_map.begin();
                for( ; itr != kvp_map.end(); ++itr ) {
                    properties_.set< std::string >( itr->first, itr->second );
                } // for itr

                uint32_t retry_attempts = irods::DEFAULT_RETRY_ATTEMPTS;
                if ( kvp_map.find( irods::RETRY_ATTEMPTS_KW ) != kvp_map.end() ) {
                    try {
                        // boost::lexical_cast does not raise errors on negatives when casting to unsigned
                        const int int_retry_attempts = boost::lexical_cast< int >( kvp_map[ irods::RETRY_ATTEMPTS_KW ] );
                        if ( int_retry_attempts < 0 ) {
                            irods::log( ERROR( SYS_INVALID_INPUT_PARAM,
                                           ( boost::format( "%s:%d - [%s] is < 0; using default value [%d]" ) %
                                             __FUNCTION__ % __LINE__ %
                                             irods::RETRY_ATTEMPTS_KW.c_str() %
                                             irods::DEFAULT_RETRY_ATTEMPTS ).str().c_str() ) );
                        }
                        else {
                            retry_attempts = static_cast< uint32_t >( int_retry_attempts );
                        }
                    }
                    catch ( const boost::bad_lexical_cast& ) {
                        irods::log( ERROR( SYS_INVALID_INPUT_PARAM,
                                        ( boost::format( "failed to cast [%s] to value [%s]; using default value [%d]" ) %
                                          irods::RETRY_ATTEMPTS_KW.c_str() %
                                          kvp_map[ irods::RETRY_ATTEMPTS_KW ] %
                                          irods::DEFAULT_RETRY_ATTEMPTS ).str().c_str() ) );
                    }
                }
                properties_.set< uint32_t >( irods::RETRY_ATTEMPTS_KW, retry_attempts );

                uint32_t retry_delay_in_seconds = irods::DEFAULT_RETRY_FIRST_DELAY_IN_SECONDS;
                if ( kvp_map.find( irods::RETRY_FIRST_DELAY_IN_SECONDS_KW ) != kvp_map.end() ) {
                    try {
                        // boost::lexical_cast does not raise errors on negatives when casting to unsigned
                        const int int_retry_delay = boost::lexical_cast< int >( kvp_map[ irods::RETRY_FIRST_DELAY_IN_SECONDS_KW ] );
                        if ( int_retry_delay <= 0 ) {
                            irods::log( ERROR( SYS_INVALID_INPUT_PARAM,
                                            ( boost::format( "%s:%d - [%s] is <= 0; using default value [%d]" ) %
                                              __FUNCTION__ % __LINE__ %
                                              irods::RETRY_FIRST_DELAY_IN_SECONDS_KW.c_str() %
                                              irods::DEFAULT_RETRY_FIRST_DELAY_IN_SECONDS ).str().c_str() ) );
                        }
                        else {
                            retry_delay_in_seconds = static_cast< uint32_t >( int_retry_delay );
                        }
                    }
                    catch ( const boost::bad_lexical_cast& ) {
                        irods::log( ERROR( SYS_INVALID_INPUT_PARAM,
                                        ( boost::format( "failed to cast [%s] to value [%s]; using default value [%d]" ) %
                                          irods::RETRY_BACKOFF_MULTIPLIER_KW.c_str() %
                                          kvp_map[ irods::RETRY_FIRST_DELAY_IN_SECONDS_KW ] %
                                          irods::DEFAULT_RETRY_FIRST_DELAY_IN_SECONDS ).str().c_str() ) );
                    }
                }
                properties_.set< uint32_t >( irods::RETRY_FIRST_DELAY_IN_SECONDS_KW, retry_delay_in_seconds );

                double backoff_multiplier = irods::DEFAULT_RETRY_BACKOFF_MULTIPLIER;
                if ( kvp_map.find( irods::RETRY_BACKOFF_MULTIPLIER_KW ) != kvp_map.end() ) {
                    try {
                        backoff_multiplier = boost::lexical_cast< double >( kvp_map[ irods::RETRY_BACKOFF_MULTIPLIER_KW ] );
                        if ( backoff_multiplier < 1 ) {
                            irods::log( ERROR( SYS_INVALID_INPUT_PARAM,
                                            ( boost::format( "%s:%d - [%s] is < 1; using default value [%f]" ) %
                                              __FUNCTION__ % __LINE__ %
                                              irods::RETRY_BACKOFF_MULTIPLIER_KW.c_str() %
                                              irods::DEFAULT_RETRY_BACKOFF_MULTIPLIER ).str().c_str() ) );
                            backoff_multiplier = irods::DEFAULT_RETRY_BACKOFF_MULTIPLIER;
                        }
                    }
                    catch ( const boost::bad_lexical_cast& ) {
                        irods::log( ERROR( SYS_INVALID_INPUT_PARAM,
                                        ( boost::format( "failed to cast [%s] to value [%s]; using default value [%f]" ) %
                                          irods::RETRY_BACKOFF_MULTIPLIER_KW.c_str() %
                                          kvp_map[ irods::RETRY_BACKOFF_MULTIPLIER_KW ] %
                                          irods::DEFAULT_RETRY_BACKOFF_MULTIPLIER ).str().c_str() ) );
                    }
                }
                properties_.set< double >( irods::RETRY_BACKOFF_MULTIPLIER_KW, backoff_multiplier );

            } // ctor

            irods::error post_disconnect_maintenance_operation(
                irods::pdmo_type& ) {
                irods::error result = SUCCESS();
                // nothing to do
                return result;
            }

            irods::error need_post_disconnect_maintenance_operation(
                bool& _flag ) {
                irods::error result = SUCCESS();
                _flag = false;
                return result;
            }

    }; // class repl_resource

    // =-=-=-=-=-=-=-
    // 4. create the plugin factory function which will return a dynamically
    //    instantiated object of the previously defined derived resource.  use
    //    the add_operation member to associate a 'call name' to the interfaces
    //    defined above.  for resource plugins these call names are standardized
    //    as used by the irods facing interface defined in
    //    server/drivers/src/fileDriver.c
    irods::resource* plugin_factory( const std::string& _inst_name, const std::string& _context ) {

        // =-=-=-=-=-=-=-
        // 4a. create repl_resource
        repl_resource* resc = new repl_resource( _inst_name, _context );

        // =-=-=-=-=-=-=-
        // 4b. map function names to operations.  this map will be used to load
        //     the symbols from the shared object in the delay_load stage of
        //     plugin loading.
        resc->add_operation( irods::RESOURCE_OP_CREATE,            "replFileCreate" );
        resc->add_operation( irods::RESOURCE_OP_OPEN,              "replFileOpen" );
        resc->add_operation( irods::RESOURCE_OP_READ,              "replFileRead" );
        resc->add_operation( irods::RESOURCE_OP_WRITE,             "replFileWrite" );
        resc->add_operation( irods::RESOURCE_OP_CLOSE,             "replFileClose" );
        resc->add_operation( irods::RESOURCE_OP_UNLINK,            "replFileUnlink" );
        resc->add_operation( irods::RESOURCE_OP_STAT,              "replFileStat" );
        resc->add_operation( irods::RESOURCE_OP_MKDIR,             "replFileMkdir" );
        resc->add_operation( irods::RESOURCE_OP_OPENDIR,           "replFileOpendir" );
        resc->add_operation( irods::RESOURCE_OP_READDIR,           "replFileReaddir" );
        resc->add_operation( irods::RESOURCE_OP_RENAME,            "replFileRename" );
        resc->add_operation( irods::RESOURCE_OP_TRUNCATE,          "replFileTruncate" );
        resc->add_operation( irods::RESOURCE_OP_FREESPACE,         "replFileGetFsFreeSpace" );
        resc->add_operation( irods::RESOURCE_OP_LSEEK,             "replFileLseek" );
        resc->add_operation( irods::RESOURCE_OP_RMDIR,             "replFileRmdir" );
        resc->add_operation( irods::RESOURCE_OP_CLOSEDIR,          "replFileClosedir" );
        resc->add_operation( irods::RESOURCE_OP_STAGETOCACHE,      "replStageToCache" );
        resc->add_operation( irods::RESOURCE_OP_SYNCTOARCH,        "replSyncToArch" );
        resc->add_operation( irods::RESOURCE_OP_RESOLVE_RESC_HIER, "replRedirect" );
        resc->add_operation( irods::RESOURCE_OP_REGISTERED,        "replFileRegistered" );
        resc->add_operation( irods::RESOURCE_OP_UNREGISTERED,      "replFileUnregistered" );
        resc->add_operation( irods::RESOURCE_OP_MODIFIED,          "replFileModified" );
        resc->add_operation( irods::RESOURCE_OP_REBALANCE,         "replRebalance" );
        resc->add_operation( irods::RESOURCE_OP_NOTIFY,            "replFileNotify" );

        // =-=-=-=-=-=-=-
        // set some properties necessary for backporting to iRODS legacy code
        resc->set_property< int >( irods::RESOURCE_CHECK_PATH_PERM, 2 );//DO_CHK_PATH_PERM );
        resc->set_property< int >( irods::RESOURCE_CREATE_PATH,     1 );//CREATE_PATH );

        // =-=-=-=-=-=-=-
        // 4c. return the pointer through the generic interface of an
        //     irods::resource pointer
        return dynamic_cast<irods::resource*>( resc );

    } // plugin_factory

}; // extern "C"
